import QtQuick 2.1
import QtQuick.Controls 1.1
import QtQuick.Controls.Styles 1.1
import QtQuick.Layouts 1.1
import DemoPlugin 1.0

// TODO: 
//   guiding centre/epicycles
//   axes/plane only extend as far as orbit
//   indication of where these are in MW
//   rotating frame
//   sounds
//   Ralphification
//   Multi orbits / time-avgd galaxy building

Rectangle {
    id : subroot
    opacity: 1.0
    color: "#00FFFFFF"
    property int nframes : 10000
    property real lognframes : Math.log(nframes*1.0)/Math.log(10.0)
    property real dtfac : 1e-2
    property real x0 : 2.5 // ICs
    property real vy : 0.4
    property real vz : 0.2
    property real mass : 1.0
    property real q : 0.8
    
    // This is the orbit/potential we start with.
    // The definition of what an 'Orbit' is in the c++ file
    // plugin/orbit.cpp.  This particular 'Orbit' instance is
    // referred to as 'myorbit' in the QML code below.
    // In particular, it's projected onto the screen background
    // by feeding  it in to an 'OrbitWidget'
    Orbit {
        id : myorbit
        mass : subroot.mass   // characteristic mass of potential
        q  : subroot.q     // flattening of potential
        x0 : Qt.vector3d(subroot.x0, 0.0, 0.0) // ICs
        v0 : Qt.vector3d(0.0, subroot.vy, subroot.vz)
    }    
    // Declare that the whole Rectangle is part of a PinchArea
    // and MouseArea.  By pinching or dragging inside this area
    // we can control the projection of the orbit.
    PinchArea {
        anchors.fill: parent
        onPinchUpdated: {
            // Rotate/zoom if pinched
            var m = myorbitwidget.modelview
            var f = pinch.previousScale/pinch.scale
            var dangle = (pinch.angle - pinch.previousAngle)*Math.PI/180
            var cosphi = Math.cos(dangle), sinphi = Math.sin(dangle)
            var mrotscale = Qt.matrix4x4( cosphi, -sinphi, 0,0,
                                          sinphi, cosphi, 0,0,
                                          0,            0, 1,0,
                                          0,            0, 0,f)
            myorbitwidget.modelview = mrotscale.times(m)
        }
        MouseArea {
            anchors.fill: parent
            hoverEnabled: false
            property int previousX: 0
            property int previousY: 0
            onPressed: {
                previousX = mouse.x
                previousY = mouse.y
            }
            onPositionChanged: {
                // Change orientation if dragged
                var bx = mouse.x - previousX
                var by = mouse.y - previousY
                previousX = mouse.x
                previousY = mouse.y
                var phix = by*0.01
                var phiy = bx*0.01
                var  m = myorbitwidget.modelview
                var cosphix = Math.cos(phix), sinphix = Math.sin(phix);
                var Rx = Qt.matrix4x4( 1,      0,      0, 0,
                                       0, cosphix,-sinphix, 0,
                                       0, sinphix, cosphix, 0,
                                       0,      0,      0, 1)
                var cosphiy = Math.cos(phiy), sinphiy = Math.sin(phiy);
                var Ry = Qt.matrix4x4( cosphiy, 0,  sinphiy, 0,
                                       0,       1,        0, 0,
                                       -sinphiy,0,  cosphiy, 0,
                                       0,       0,        0, 1)
                myorbitwidget.modelview = Rx.times(Ry.times(m))
            }

            // An 'OrbitWidget' projects an 'Orbit' onto
            // the *background*.  The implementation of
            // 'OrbitWidget' is in the c++ file plugin/orbitwidget.cpp.
            // The definition of what an 'Orbit' is in the c++ file
            // This particular OrbitWidget is referred to as
            // 'myorbitwidget'.  The 'modelview' property directly
            // controls the projection of 'myorbit' onto the screen.
            OrbitWidget {
                id : myorbitwidget
                orbit : myorbit
                modelview : Qt.matrix4x4(0.1,0,0,0,
                                         0,0.1,0,0,
                                         0,0,0.1,0,
                                         0,0,0,1)
                // Animate the orbit by making the 'frac2Show'
                // member of the c++ OrbitWidget class run
                // continuously between 0 and 1.
                property bool animated : true
                SequentialAnimation on frac2Show {
                    id: animation
                    NumberAnimation {
                        from: 0.0;
                        to: 1.0;
                        duration: 5000;
                        //easing.type: Easing.InQuad
                    }
                    PauseAnimation { duration: 1000; }
                    loops: Animation.Infinite
                    running: true
                }
            }
        }
    }
    Text {
        color: "white"
        font.pixelSize: blsize;
        anchors.top: parent.top
        anchors.left: parent.left
        property real time: (myorbitwidget.frac2Show*nframes*dtfac*x0*10).toFixed(0)
        text: "Time: "+time+" Myr"
        anchors.margins: 20
    }
        
    // Now we set up the controls.
    // This should make sense after a few readings.
    Column {
        id : controls_top
        anchors.right: parent.right
        anchors.top: parent.top
        anchors.margins: 30
        spacing : 0.5*blsize
        z: 3
        Row {
            spacing: 10
            Switch {
                z: 20
                style: switchStyle
                checked: myorbitwidget.showAxes
                onCheckedChanged: {
                    myorbitwidget.showAxes = !(myorbitwidget.showAxes)
                }
            }
            Text { text: "Axes"; font.pixelSize: 0.8*blsize; color: "white" }
        }
        Row {
            spacing: 10
            Switch {
                z: 20
                style: switchStyle
                checked: myorbitwidget.animated
                onCheckedChanged: { /// Eh? Wrong way round?
                    if(myorbitwidget.animated) {
                        myorbitwidget.animated = false;
                        animation.resume()
                    } else {
                        myorbitwidget.animated = true;
                        animation.pause()
                    }
                }
            }
            Text {
                text: "Animation";
                font.pixelSize: 0.8*blsize; color: "white"
            }
        }
        Row {
            spacing: 10
            Slider {
                z: 20
                style: touchStyle
                minimumValue: 3
                maximumValue: 5
                stepSize: 0.1
                tickmarksEnabled: true
                orientation: Qt.Horizontal
                updateValueWhileDragging: false
                value: lognframes
                onValueChanged: {
                    lognframes = value
                    nframes = Math.pow(10.0,lognframes)
                    myorbit.x0 = Qt.vector3d(x0,0,0)
                    myorbit.v0 = Qt.vector3d(0,vy,vz)
                    myorbit.leapfrog(dtfac*x0,nframes)
                    myorbitwidget.orbit = myorbit
                }
            }
            Text { text: "length";
                   font.pixelSize: 0.8*blsize; color: "white" }
        }
    }
    Column {
        id: controls_ics
        anchors.right: parent.right
        anchors.top: controls_top.bottom
        anchors.left: controls_top.left
        anchors.margins: 30
        spacing : 0.5*blsize
        z: 3
        Text { text: "<b>Launch from</b>";
               font.pixelSize: blsize; color: "white"
             }
        Row {
            spacing: 0
            Text { text: "Horiz. position:";
                   font.pixelSize: 0.8*blsize; color: "white" }
        }
        Row {
            spacing: 0
            Slider {
                z: 20
                style: touchStyle
                minimumValue: 0
                maximumValue: 5
                value: x0
                orientation: Qt.Horizontal
                updateValueWhileDragging: true
                onValueChanged: {
                    x0 = value
                    myorbit.x0 = Qt.vector3d(x0,0,0)
                    myorbit.v0 = Qt.vector3d(0,vy,vz)
                    myorbit.leapfrog(dtfac*x0,nframes)
                    myorbitwidget.orbit = myorbit
                }
            }
        }
        Row { spacing: blsize }
        Row {
            Text { text: "Horiz. speed:";
                   font.pixelSize: 0.8*blsize; color: "white" }
        }
        Row {
            Slider {
                z: 20
                style: touchStyle
                minimumValue: 0
                maximumValue: 1
                value: vy
                orientation: Qt.Horizontal
                updateValueWhileDragging: true
                onValueChanged: {
                    vy = value
                    myorbit.x0 = Qt.vector3d(x0,0,0)
                    myorbit.v0 = Qt.vector3d(0,vy,vz)
                    myorbit.leapfrog(dtfac*x0,nframes)
                    myorbitwidget.orbit = myorbit
                }
            }
        }
        Row { spacing: blsize }
        Row {
            Text { text: "Vert. speed:";
                   font.pixelSize: 0.8*blsize; color: "white" }
        }
        Row{
            Slider {
                z: 20
                style: touchStyle
                minimumValue: 0
                maximumValue: 1
                orientation: Qt.Horizontal
                updateValueWhileDragging: true
                value: vz
                onValueChanged: {
                    vz = value
                    myorbit.x0 = Qt.vector3d(x0,0,0)
                    myorbit.v0 = Qt.vector3d(0,vy,vz)
                    myorbit.leapfrog(dtfac*x0,nframes)
                    myorbitwidget.orbit = myorbit
                }
            }
        }
            
    }

    Column {
        id: controls_pot
        anchors.right: parent.right
        anchors.top: controls_ics.bottom
        anchors.left: controls_top.left
        anchors.margins: 30
        spacing : 0.5*blsize
        z: 3
        Text { text: "<b>Galaxy model</b>"
               font.pixelSize: blsize; color: "white"
             }
        Row {
            spacing: 0
            Text { text: "Mass:";
                   font.pixelSize: 0.8*blsize; color: "white" }
        }
        Row {
            spacing: 0
            Slider {
                z: 20
                style: touchStyle
                minimumValue: 0
                maximumValue: 2
                value: mass
                orientation: Qt.Horizontal
                updateValueWhileDragging: true
                onValueChanged: {
                    mass = value
                    myorbit.mass = value
                    myorbit.x0 = Qt.vector3d(x0,0,0)
                    myorbit.v0 = Qt.vector3d(0,vy,vz)
                    myorbit.leapfrog(dtfac*x0,nframes)
                    myorbitwidget.orbit = myorbit
                }
            }
        }
        Row { spacing: blsize }
        Row {
            Text { text: "Flattening:";
                   font.pixelSize: 0.8*blsize; color: "white" }
        }
        Row {
            Slider {
                z: 20
                style: touchStyle
                minimumValue: 0.3
                maximumValue: 1.5
                value: q
                orientation: Qt.Horizontal
                updateValueWhileDragging: true
                onValueChanged: {
                    q = value
                    myorbit.x0 = Qt.vector3d(x0,0,0)
                    myorbit.v0 = Qt.vector3d(0,vy,vz)
                    myorbit.leapfrog(dtfac*x0,nframes)
                    myorbitwidget.orbit = myorbit
                }
            }
        }
    }
    // Prevent mouse events within control rectangles
    // from affecting modelview
    Rectangle{ color: "#00000000";
               anchors.fill: controls_top;
               z: 1;
               MouseArea { anchors.fill: parent; } }
    Rectangle{ color: "#00000000";
               anchors.fill: controls_ics;
               z: 1;
               MouseArea { anchors.fill: parent; } }
    Rectangle{ color: "#00000000";
               anchors.fill: controls_pot;
               z: 1;
               MouseArea { anchors.fill: parent; } }

    Rectangle {
        color: Qt.rgba(0.2, 0.2, 0.2, 0.3)
        radius: 10
        border.width: 1
        border.color: "white"
        anchors.fill: controls_top
        anchors.margins: -10
    }        
    Rectangle {
        color: Qt.rgba(0.2, 0.2, 0.2, 0.3)
        radius: 10
        border.width: 1
        border.color: "white"
        anchors.fill: controls_ics
        anchors.margins: -10
    }        
    Rectangle {
        color: Qt.rgba(0.2, 0.2, 0.2, 0.3)
        radius: 10
        border.width: 1
        border.color: "white"
        anchors.fill: controls_pot
        anchors.margins: -10
    }        


    Rectangle {
        color: Qt.rgba(1, 1, 1, 0.7)
        radius: 10
        border.width: 1
        border.color: "white"
        anchors.fill: label
        anchors.margins: -10
    }    
    Text {
        id: label
        color: "black"
        font.pixelSize: blsize
        wrapMode: Text.WordWrap
        property bool hidden: false
        text: hidden ?  " ? " : "The <b>orbit</b> of a star is just the path it follows as it moves around the Galaxy.
<br />
Like everything else in the universe, each star `wants' to move in a
straight line.  But the gravitational attraction of all the other stars
(and dark matter) pulls it inwards, making its orbit curved.
<br/>
Use the sliders to investigate orbits in different simulated galaxies.
"
        anchors.right: controls_pot.left
        anchors.left: parent.left
        anchors.bottom: parent.bottom
        anchors.margins: 30
        function toggle() {
            hidden = !hidden;
            if (hidden) state = "hidden"
            else state = "unhidden"
        }
        states: [
            State {
                name: "hidden"
                AnchorChanges {
                    target: label
                    anchors.right: hidden_question.right //controls_pot.right
                }
                PropertyChanges {
                    target: label
                    //anchors.topMargin: 10
                    //anchors.bottomMargin: 10
                }
            },
            State {
                name: "unhidden"
                AnchorChanges {
                    target: label
                    anchors.right: controls_pot.left
                }
                PropertyChanges {
                    target: label
                    //anchors.topMargin: 10
                    //anchors.bottomMargin: 10
                }
            }
        ]

    }
    MouseArea {
        anchors.fill: label
        onClicked: { label.toggle() }
    }
    Text { // stupid stupid stupid
        id: hidden_question
        anchors.left: label.left
        anchors.top: label.top
        anchors.margins: 30
        font.pixelSize: blsize
        color: "#00000000"
        text: " ? "
    }
        
    Component {
        id: switchStyle
        SwitchStyle {

            groove: Rectangle {
                //implicitHeight: 50
                //implicitWidth: 152
                implicitHeight: 30
                implicitWidth: 102
                Rectangle {
                    anchors.top: parent.top
                    anchors.left: parent.left
                    anchors.bottom: parent.bottom
                    width: parent.width/2 - 2
                    height: blsize
                    anchors.margins: 2
                    color: control.checked ? "#468bb7" : "#222"
                    Behavior on color {ColorAnimation {}}
                    Text {
                        font.pixelSize: blsize*0.6
                        color: "white"
                        anchors.centerIn: parent
                        text: "ON"
                    }
                }
                Item {
                    width: parent.width/2
                    height: parent.height
                    anchors.right: parent.right
                    Text {
                        font.pixelSize: blsize*0.6
                        color: "white"
                        anchors.centerIn: parent
                        text: "OFF"
                    }
                }
                color: "#222"
                border.color: "#444"
                border.width: 2
            }
            handle: Rectangle {
                width: parent.parent.width/2
                height: control.height
                color: "#444"
                border.color: "#555"
                border.width: 2
            }
        }
    }
    Component {
        id: touchStyle
        SliderStyle {
            handle: Rectangle {
                width: 1.5*blsize
                height: 1.5*blsize
                radius: height*0.4
                antialiasing: true
                color: Qt.lighter("#468bb7", 1.2)
            }

            groove: Item {
                implicitHeight: blsize
                implicitWidth: 8*blsize
                Rectangle {
                    height: 0.3*blsize
                    width: parent.width
                    anchors.verticalCenter: parent.verticalCenter
                    color: "#444"
                    opacity: 0.8
                    Rectangle {
                        antialiasing: true
                        radius: 1
                        color: "#468bb7"
                        height: parent.height
                        width: parent.width * (control.value-control.minimumValue) / (control.maximumValue-control.minimumValue)
                    }
                }
            }
        }
    }
}
