#include <iostream>
using namespace std;

#include "orbit.h"
#include "orbitcloud.h"
#include <QVector>
#include <QVector3D>

#include "orbitcloudwidget.h"

#include <QtQuick/qquickwindow.h>
#include <QtGui/QOpenGLShaderProgram>
#include <QtGui/QOpenGLContext>

OrbitCloudWidget::OrbitCloudWidget(QQuickItem *parent)
  : QQuickItem(parent)
  , m_program(0)
  , m_modelview(QMatrix4x4())
  , m_thread_modelview(QMatrix4x4())
  , m_frac2show(-1)
  , m_thread_frac2show(-1)
  , vertexAttr1(0)
  , matrixUniform1(0)
  , N(0)
  , vertices(0)
  , m_orbits(0)
{
    connect(this, SIGNAL(windowChanged(QQuickWindow*)), this, SLOT(handleWindowChanged(QQuickWindow*)));
}

void OrbitCloudWidget::setFrac2Show(qreal frac) {
  if (frac == m_frac2show) return;
    m_frac2show = frac;
    emit frac2ShowChanged();
    if (window())   window()->update();
}

void OrbitCloudWidget::setModelview(QMatrix4x4 modelview) {
  if(modelview == m_modelview) return;
  m_modelview = modelview;
  if(window()) window()->update();
}
void OrbitCloudWidget::setOrbit(OrbitCloud *orbits) {
  m_orbits = orbits;
  N = orbits->xs.size();
  if(vertices) { delete[] vertices; vertices = 0; }
  vertices = new GLfloat[3*N];
  for(int n=0;n<N;n++)
    for(int i=0;i<3;i++)
      vertices[3*n+i] = orbits->xs[n][i];
  m_frac2show = -1;
  emit orbitChanged();
    if (window()) window()->update();
}
void OrbitCloudWidget::setShowAxes(bool onoff) {
  if(m_showaxes == onoff) return;
  m_showaxes = onoff;
  if(window()) window()->update();
}

void OrbitCloudWidget::handleWindowChanged(QQuickWindow *win)
{
    if (win) {
        // Connect the beforeRendering signal to our paint function.
        // Since this call is executed on the rendering thread it must be
        // a Qt::DirectConnection
        connect(win, SIGNAL(beforeRendering()), this, SLOT(paint()), Qt::DirectConnection);
        connect(win, SIGNAL(beforeSynchronizing()), this, SLOT(sync()), Qt::DirectConnection);

        // If we allow QML to do the clearing, they would clear what we paint
        // and nothing would show.
        win->setClearBeforeRendering(false);
    }
}

void OrbitCloudWidget::paint() {
    qreal ratio = window()->devicePixelRatio();
    int w = int(ratio * window()->width());
    int h = int(ratio * window()->height());
    glViewport(0,0, w, h);
    double aspect = w*1.0/h;

    if(!m_program) {
      m_program = new QOpenGLShaderProgram();
      m_program->addShaderFromSourceCode(QOpenGLShader::Vertex,
                                         "attribute highp vec4 vertex;"
                                         "uniform mediump mat4 P;"
                                         "uniform mediump mat4 aspect;"
                                         "uniform mediump vec4 incolor;"
                                         "varying mediump vec4 color;"
                                         "void main() {"
                                         "    color = incolor;"
                                         "    gl_Position = aspect*(P*vertex);"
                                         "    gl_PointSize = 1.0;"
                                          "}");
      m_program->addShaderFromSourceCode(QOpenGLShader::Fragment,
                                          "varying mediump vec4 color;"
                                          "void main() {"
                                          "  gl_FragColor = color;"
                                          "}");
      m_program->bindAttributeLocation("vertex", 0);
      m_program->link();
    }
    m_program->bind();
    m_program->enableAttributeArray("vertex");
    
    QMatrix4x4 amatrix = QMatrix4x4();
    amatrix(1,1) = aspect;
    m_program->setUniformValue("aspect",amatrix);
    m_program->setUniformValue( "P",
                                m_thread_modelview);

    glDisable(GL_DEPTH_TEST);

    glClearColor(0.1f, 0.1f, 0.1f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT);

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE);

    // orbit
    int num2plot = m_thread_frac2show < 0 ? N/3 : int(N/3*m_thread_frac2show);

    for(int i=0;i<3;i++){
        m_program->setAttributeArray("vertex",
                                     GL_FLOAT, vertices, 3);
        m_program->setUniformValue("incolor",
                                    QVector4D(0.2,1.0,0.2,1.0) );
      glDrawArrays(GL_LINE_STRIP, i*N/3, num2plot);

        // IC marker.  Could easily make this a tetrahedron
        // or something.
        float ic_x = vertices[i*N+0], ic_y = vertices[i*N+1], ic_z = vertices[i*N+2];
        float dx = 0.03,dy = dx;
        float icvertices[] = {
          ic_x-dx, ic_y-dy, ic_z,
          ic_x-dx, ic_y+dy, ic_z,
          ic_x+dx, ic_y, ic_z};
        m_program->setAttributeArray("vertex",
                                     GL_FLOAT, icvertices, 3);
        m_program->setUniformValue("incolor",
                                    QVector4D(1.0,1.0,0.0, 1.0) );
        glDrawArrays(GL_TRIANGLE_STRIP, 0, 3);

        float icvertices2[] = {
          ic_x+0.5*dx, ic_y+dy, ic_z,
          ic_x+0.5*dx, ic_y-dy, ic_z,
          ic_x-1.5*dx, ic_y, ic_z};
        m_program->setAttributeArray("vertex",
                                     GL_FLOAT, icvertices2, 3);
        m_program->setUniformValue("incolor",
                                    QVector4D(1.0,1.0,0.0, 1.0) );
        glDrawArrays(GL_TRIANGLE_STRIP, 0, 3);
    }
    if(m_thread_showaxes) {
      // equatorial plane
      float eqvertices[] = { -10,-10,0,
                             +10,-10,0,
                             -10,+10,0,
                             +10,+10,0 };
      m_program->setAttributeArray("vertex", GL_FLOAT, eqvertices, 3);
      m_program->setUniformValue("incolor",
                                 QVector4D(0.8,0.1,0.1,0.2) );
      glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
      // coord axes
      float axvertices[] = { 0,0,-10, 0,0,+10,
                             0,-10,0, 0,+10,0,
                             -10,0,0, +10,0,0};
      m_program->setAttributeArray("vertex", GL_FLOAT, axvertices, 3);
      m_program->setUniformValue("incolor",
                                 QVector4D(1.0,1.0,1.0,1.0) );
      glDrawArrays(GL_LINES, 0, 6);
    }
    

    m_program->disableAttributeArray("vertex");
    m_program->release();
}

void OrbitCloudWidget::cleanup() {
    if (m_program) {
        delete m_program;
        m_program = 0;
    }
    if(vertices) {
      delete[] vertices;
      vertices = 0;
    }
}

void OrbitCloudWidget::sync() {
  m_thread_modelview = m_modelview;
  m_thread_frac2show = m_frac2show;
  m_thread_showaxes  = m_showaxes;
}

