#include <iostream>
#include <math.h>
using namespace std;

#include "orbit.h"
#include <QVector>
#include <QVector3D>

#include "orbitwidget.h"

#include <QtQuick/qquickwindow.h>
#include <QtGui/QOpenGLShaderProgram>
#include <QtGui/QOpenGLContext>

OrbitWidget::OrbitWidget(QQuickItem *parent)
  : QQuickItem(parent)
  , m_program(0)
  , m_modelview(QMatrix4x4())
  , m_thread_modelview(QMatrix4x4())
  , m_frac2show(-1)
  , m_thread_frac2show(-1)
  , N(0)
  , m_orbit(0)
{
    connect(this, SIGNAL(windowChanged(QQuickWindow*)), this, SLOT(handleWindowChanged(QQuickWindow*)));
}

OrbitWidget::~OrbitWidget() {
  cleanup();
}

void OrbitWidget::setFrac2Show(qreal frac) {
  if (frac == m_frac2show) return;
    m_frac2show = frac;
    emit frac2ShowChanged();
    if (window())   window()->update();
}

void OrbitWidget::setModelview(QMatrix4x4 modelview) {
  if(modelview == m_modelview) return;
  m_modelview = modelview;
  if(window()) window()->update();
}
void OrbitWidget::setOrbit(Orbit *orbit) {
  m_orbit = orbit;
  if(N!=(orbit->nstep)) {
    N = min(maxnstep,orbit->nstep);
  }
  xmax = 0.0;
  for(int n=0;n<N;n++)
    for(int i=0;i<3;i++) {
      vertices[3*n+i] = orbit->xs[n][i];
      xmax = max(xmax,(qreal) fabs(orbit->xs[n][i]));
    }
  m_frac2show = -1;
  emit orbitChanged();
  if (window()) window()->update();
}
void OrbitWidget::setShowAxes(bool onoff) {
  if(m_showaxes == onoff) return;
  m_showaxes = onoff;
  if(window()) window()->update();
}

void OrbitWidget::handleWindowChanged(QQuickWindow *win)
{
    if (win) {
        // Connect the beforeRendering signal to our paint function.
        // Since this call is executed on the rendering thread it must be
        // a Qt::DirectConnection
        connect(win, SIGNAL(beforeRendering()), this, SLOT(paint()), Qt::DirectConnection);
        connect(win, SIGNAL(beforeSynchronizing()), this, SLOT(sync()), Qt::DirectConnection);

        // If we allow QML to do the clearing, they would clear what we paint
        // and nothing would show.
        win->setClearBeforeRendering(false);
    }
}

void OrbitWidget::plotEqPlane() {
  float rmax = xmax*1.2;
  // equatorial plane
  float eqvertices[] = { -rmax,-rmax,0,
                         +rmax,-rmax,0,
                         -rmax,+rmax,0,
                         +rmax,+rmax,0 };
  m_program->setAttributeArray("vertex", GL_FLOAT, eqvertices, 3);
  m_program->setUniformValue("incolor",
                             QVector4D(0.8,0.1,0.1,0.2) );
  glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
  // coord axes
  rmax *= 1.1;
  float axvertices[] = { 0,0,-rmax, 0,0,+rmax,
                         0,-rmax,0, 0,+rmax,0,
                         -rmax,0,0, +rmax,0,0};
  m_program->setAttributeArray("vertex", GL_FLOAT, axvertices, 3);
  m_program->setUniformValue("incolor",
                             QVector4D(1.0,1.0,1.0,1.0) );
  glLineWidth(3);
  glDrawArrays(GL_LINES, 0, 6);
}

void OrbitWidget::plotTetrahedron(const QVector3D &x,
                                  const QVector4D &color,
                                  float dx) {
  float ic_x = x[0], ic_y = x[1], ic_z = x[2];
  float dy = dx, dz = dx;
  float icvertices[] = {
    // Front face
    -1.0f, -1.0f,  1.0f,  
    1.0f, -1.0f,  1.0f,  
    1.0f,  1.0f,  1.0f,  
    //Right face
    -1.0f, -1.0f,  1.0f,  
    1.0f,  1.0f, 1.0f,
    -1.0f, -1.0f, -1.0f, 
    // Left Face
    1.0f, -1.0f, 1.0f,  
    -1.0f, -1.0f, -1.0f, 
    1.0f,  1.0f, 1.0f,  
    // bottom
    -1.0f, -1.0f, 1.0f,  
    -1.0f, -1.0f, -1.0f, 
    1.0f, -1.0f,  1.0f  };
  // Scale and translate
  for(int i=0;i<12;i++) {
    icvertices[3*i+0] = ic_x + dx*icvertices[3*i+0];
    icvertices[3*i+1] = ic_y + dy*icvertices[3*i+1];
    icvertices[3*i+2] = ic_z + dz*icvertices[3*i+2];
  }
  m_program->setAttributeArray("vertex",
                               GL_FLOAT, icvertices, 3);
  m_program->setUniformValue("incolor",color);
  glDrawArrays(GL_TRIANGLE_STRIP, 0, 3);
  m_program->setUniformValue("incolor",color);
  glDrawArrays(GL_TRIANGLE_STRIP, 3, 3); // 3 vertices from posn 3
  m_program->setUniformValue("incolor",color);
  glDrawArrays(GL_TRIANGLE_STRIP, 6, 3); // 3 vertices from posn 6
  m_program->setUniformValue("incolor",color);
  glDrawArrays(GL_TRIANGLE_STRIP, 9, 3); // 3 vertices from posn 9
}

void OrbitWidget::paint() {
    qreal ratio = window()->devicePixelRatio();
    int w = int(ratio * window()->width());
    int h = int(ratio * window()->height());
    glViewport(0,0, w, h);
    double aspect = w*1.0/h;

    if(!m_program) {
      m_program = new QOpenGLShaderProgram();
      m_program->addShaderFromSourceCode(QOpenGLShader::Vertex,
                                         "attribute highp vec4 vertex;"
                                         "uniform mediump mat4 P;"
                                         "uniform mediump vec4 incolor;"
                                         "varying mediump vec4 color;"
                                         "void main() {"
                                         "    color = incolor;"
                                         "    gl_Position = P*vertex;"
                                         "    gl_PointSize = 1.0;"
                                          "}");
      m_program->addShaderFromSourceCode(QOpenGLShader::Fragment,
                                          "varying mediump vec4 color;"
                                          "void main() {"
                                          "  gl_FragColor = color;"
                                          "}");
      m_program->bindAttributeLocation("vertex", 0);
      m_program->link();
    }
    m_program->bind();
    m_program->enableAttributeArray("vertex");
    
    QMatrix4x4 amatrix = QMatrix4x4();
    amatrix(1,1) = aspect;

    // // Can't get following to work.  Going to stick
    // // with original version above.
    // const qreal zNear = -1.0, zFar = 1.0, fov = 10.0;
    // // Reset projection
    // amatrix.setToIdentity();
    // // Set perspective projection
    // amatrix.perspective(fov, aspect, zNear, zFar);
    // for(int i=0;i<4;i++) {
    //   for(int j=0;j<4;j++) cout << amatrix(i,j) << " ";
    //   cout << endl;
    // }
    // cout << endl;
    
    m_program->setUniformValue( "P",
                                amatrix*m_thread_modelview);

    glDisable(GL_DEPTH_TEST);

    glClearColor(0.1f, 0.1f, 0.1f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT);

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE);

    // Enable depth buffer -- NEW
    //glEnable(GL_DEPTH_TEST);

    // Enable back face culling -- NEW
    // Note how with this the axes triangles become transparent
    // in one direction.  So it looks as if opengl gives
    // each triangle an orientation based on the order in which
    // the vertices are given.  Simplest way of making axes
    // double sided is to disable GL_CULL_FACE before drawing.
    // (See commented-out line later, just before axes drawing.)
    //glEnable(GL_CULL_FACE);
    
    // orbit
    m_program->setAttributeArray("vertex",
                                 GL_FLOAT, vertices, 3);
    m_program->setUniformValue("incolor",
                                QVector4D(0.2,1.0,0.2,1.0) );
    int num2plot = m_thread_frac2show < 0 ? N : int(N*m_thread_frac2show);
    glLineWidth(1);
    glDrawArrays(GL_LINE_STRIP, 0, num2plot);

    float dx = 0.05*max(max(vertices[0],vertices[1]),vertices[2]);
    plotTetrahedron(QVector3D(vertices[0],vertices[1],vertices[2]),
                    QVector4D(1.0,0,0,1),dx);
    plotTetrahedron(QVector3D(3.0,0,0),
                    QVector4D(1,1,0,1),dx);
    if(num2plot>0)
      plotTetrahedron(QVector3D(vertices[3*num2plot-3],
                                vertices[3*num2plot-2],
                                vertices[3*num2plot-1]),
                      QVector4D(1.0,0,0,1), dx);

    //glDisable(GL_CULL_FACE);
    if(m_thread_showaxes) {
      plotEqPlane();
    }
    

    m_program->disableAttributeArray("vertex");
    m_program->release();
}

void OrbitWidget::cleanup() {
    if (m_program) {
        delete m_program;
        m_program = 0;
    }
}

void OrbitWidget::sync() {
  m_thread_modelview = m_modelview;
  m_thread_frac2show = m_frac2show;
  m_thread_showaxes  = m_showaxes;
}

